@@ -1,4 +1,4 @@
-/*	$OpenBSD: tcpbench.c,v 1.59 2018/09/28 19:01:52 bluhm Exp $	*/
+/*	$OpenBSD: tcpbench.c,v 1.69 2023/05/22 12:53:04 tobhe Exp $	*/
 
 /*
  * Copyright (c) 2008 Damien Miller <djm@mindrot.org>
@@ -54,16 +54,11 @@
 #endif
 #include <netdb.h>
 #include <signal.h>
-#include <time.h>
 #include <err.h>
 #include <fcntl.h>
 #include <poll.h>
 #include <paths.h>
-
-#ifdef __OpenBSD__
-#include <kvm.h>
-#include <nlist.h>
-#endif
+#include <math.h>
 
 #define DEFAULT_PORT "12345"
 #define DEFAULT_STATS_INTERVAL 1000 /* ms */
@@ -75,6 +70,7 @@
 
 /* Our tcpbench globals */
 struct {
+	int	  Dflag;	/* Socket debug */
 	int	  Sflag;	/* Socket buffer size */
 	u_int	  rflag;	/* Report rate (ms) */
 	int	  sflag;	/* True if server */
@@ -83,11 +79,7 @@ struct {
 	int	  uflag;	/* UDP mode */
 	int	  Uflag;	/* UNIX (AF_LOCAL) mode */
 	int	  Rflag;	/* randomize client write size */
-#ifdef __OpenBSD__
-	kvm_t	 *kvmh;		/* Kvm handler */
 	char	**kvars;	/* Kvm enabled vars */
-	u_long	  ktcbtab;	/* Ktcb */
-#endif
 	char	 *dummybuf;	/* IO buffer */
 	size_t	  dummybuf_len;	/* IO buffer len */
 } tcpbench, *ptb;
@@ -109,48 +101,38 @@ struct statctx {
 	struct event ev;
 	/* TCP only */
 	struct tcpservsock *tcp_ts;
-	u_long tcp_tcbaddr;
 	/* UDP only */
 	u_long udp_slice_pkts;
 };
 
+struct statctx *udp_sc; /* singleton */
+
 static void	signal_handler(int, short, void *);
 static void	saddr_ntop(const struct sockaddr *, socklen_t, char *, size_t);
-#ifdef __OpenBSD__
-static void	drop_gid(void);
-#endif
 static void	set_slice_timer(int);
 static void	print_tcp_header(void);
-#ifdef __OpenBSD__
-static void	kget(u_long, void *, size_t);
-static u_long	kfind_tcb(int);
-static void	kupdate_stats(u_long, struct inpcb *, struct tcpcb *,
-    struct socket *);
 static void	list_kvars(void);
 static void	check_kvar(const char *);
 static char **	check_prepare_kvars(char *);
-#endif
 static void	stats_prepare(struct statctx *);
-#ifdef __OpenBSD__
+static void	summary_display(void);
 static void	tcp_stats_display(unsigned long long, long double, float,
-    struct statctx *, struct inpcb *, struct tcpcb *, struct socket *);
-#else
-static void	tcp_stats_display(unsigned long long, long double, float,
-    struct statctx *);
-#endif
+    struct statctx *, struct tcp_info *);
 static void	tcp_process_slice(int, short, void *);
 static void	tcp_server_handle_sc(int, short, void *);
 static void	tcp_server_accept(int, short, void *);
-static void	server_init(struct addrinfo *, struct statctx *);
+static void	server_init(struct addrinfo *);
 static void	client_handle_sc(int, short, void *);
-static void	client_init(struct addrinfo *, int, struct statctx *,
-    struct addrinfo *);
+static void	client_init(struct addrinfo *, int, struct addrinfo *);
 static int	clock_gettime_tv(clockid_t, struct timeval *);
 static void	udp_server_handle_sc(int, short, void *);
 static void	udp_process_slice(int, short, void *);
 #ifdef __OpenBSD__
 static int	map_tos(char *, int *);
 #endif
+static void	quit(int, short, void *);
+static void	wrapup(int);
+
 /*
  * We account the mainstats here, that is the stats
  * for all connections, all variables starting with slice
@@ -159,50 +141,81 @@ static int	map_tos(char *, int *);
  * between all slices so far.
  */
 static struct {
-	unsigned long long slice_bytes; /* bytes for last slice */
+	struct timeval t_first;		/* first connect / packet */
+ 	unsigned long long total_bytes; /* bytes since t_first */
+	unsigned long long n_slices;	/* slices since start */
+	unsigned long long slice_bytes; /* bytes since slice reset */
 	long double peak_mbps;		/* peak mbps so far */
+	long double floor_mbps;		/* floor mbps so far */
+	long double mean_mbps;		/* mean mbps so far */
+	long double nvariance_mbps;     /* for online std dev */
 	int nconns;		        /* connected clients */
 	struct event timer;		/* process timer */
+	const char *host;               /* remote server for display */
 } mainstats;
 
-#ifdef __OpenBSD__
 /* When adding variables, also add to tcp_stats_display() */
 static const char *allowed_kvars[] = {
-	"inpcb.inp_flags",
-	"sockb.so_rcv.sb_cc",
-	"sockb.so_rcv.sb_hiwat",
-	"sockb.so_rcv.sb_wat",
-	"sockb.so_snd.sb_cc",
-	"sockb.so_snd.sb_hiwat",
-	"sockb.so_snd.sb_wat",
-	"tcpcb.last_ack_sent",
-	"tcpcb.max_sndwnd",
-	"tcpcb.rcv_adv",
-	"tcpcb.rcv_nxt",
-	"tcpcb.rcv_scale",
-	"tcpcb.rcv_wnd",
-	"tcpcb.rfbuf_cnt",
-	"tcpcb.rfbuf_ts",
-	"tcpcb.snd_cwnd",
-	"tcpcb.snd_max",
-	"tcpcb.snd_nxt",
-	"tcpcb.snd_scale",
-	"tcpcb.snd_ssthresh",
-	"tcpcb.snd_una",
-	"tcpcb.snd_wl1",
-	"tcpcb.snd_wl2",
-	"tcpcb.snd_wnd",
-	"tcpcb.t_rcvtime",
-	"tcpcb.t_rtseq",
-	"tcpcb.t_rttmin",
-	"tcpcb.t_rtttime",
-	"tcpcb.t_rttvar",
-	"tcpcb.t_srtt",
-	"tcpcb.ts_recent",
-	"tcpcb.ts_recent_age",
+	"last_ack_recv",
+	"last_ack_sent",
+	"last_data_recv",
+	"last_data_sent",
+#ifdef __OpenBSD__
+	"max_sndwnd",
+#endif
+	"options",
+#ifdef __OpenBSD__
+	"rcv_adv",
+#endif
+	"rcv_mss",
+#ifdef __OpenBSD__
+	"rcv_nxt",
+	"rcv_ooopack",
+#endif
+	"rcv_space",
+#ifdef __OpenBSD__
+	"rcv_up",
+#endif
+	"rcv_wscale",
+#ifdef __OpenBSD__
+	"rfbuf_cnt",
+	"rfbuf_ts",
+#endif
+	"rtt",
+#ifdef __OpenBSD__
+	"rttmin",
+#endif
+	"rttvar",
+	"snd_cwnd",
+#ifdef __OpenBSD__
+	"snd_max",
+#endif
+	"snd_mss",
+#ifdef __OpenBSD__
+	"snd_nxt",
+	"snd_rexmitpack",
+#endif
+	"snd_ssthresh",
+#ifdef __OpenBSD__
+	"snd_una",
+	"snd_wl1",
+	"snd_wl2",
+	"snd_wnd",
+	"snd_wscale",
+	"snd_zerowin",
+	"so_rcv_sb_cc",
+	"so_rcv_sb_hiwat",
+	"so_rcv_sb_lowat",
+	"so_rcv_sb_wat",
+	"so_snd_sb_cc",
+	"so_snd_sb_hiwat",
+	"so_snd_sb_lowat",
+	"so_snd_sb_wat",
+	"ts_recent",
+	"ts_recent_age",
+#endif
 	NULL
 };
-#endif
 
 TAILQ_HEAD(, statctx) sc_queue;
 
@@ -210,13 +223,11 @@ static void
 usage(void)
 {
 	fprintf(stderr,
-#ifdef __OpenBSD__
 	    "usage: tcpbench -l\n"
-#endif
-	    "       tcpbench [-46RUuv] [-B buf] [-b addr] [-k kvars] [-n connections]\n"
+	    "       tcpbench [-46DRUuv] [-B buf] [-b sourceaddr] [-k kvars] [-n connections]\n"
 	    "                [-p port] [-r interval] [-S space] [-T toskeyword]\n"
 	    "                [-t secs] [-V rtable] hostname\n"
-	    "       tcpbench -s [-46Uuv] [-B buf] [-k kvars] [-p port] [-r interval]\n"
+	    "       tcpbench -s [-46DUuv] [-B buf] [-k kvars] [-p port] [-r interval]\n"
 	    "                [-S space] [-T toskeyword] [-V rtable] [hostname]\n");
 	exit(1);
 }
@@ -228,11 +239,20 @@ signal_handler(int sig, short event, void *bula)
 	 * signal handler rules don't apply, libevent decouples for us
 	 */
 	switch (sig) {
+#ifdef __OpenBSD__
+	case SIGINFO:
+		printf("\n");
+		wrapup(-1);
+		break;
+#endif
 	case SIGINT:
+		printf("\n");
+		wrapup(0);
+		break;		/* NOTREACHED */
 	case SIGTERM:
 	case SIGHUP:
 		warnx("Terminated by signal %d", sig);
-		exit(0);
+		wrapup(0);
 		break;		/* NOTREACHED */
 	default:
 		errx(1, "unexpected signal %d", sig);
@@ -261,18 +281,6 @@ saddr_ntop(const struct sockaddr *addr, socklen_t alen, char *buf, size_t len)
 	snprintf(buf, len, "[%s]:%s", hbuf, pbuf);
 }
 
-#ifdef __OpenBSD__
-static void
-drop_gid(void)
-{
-	gid_t gid;
-
-	gid = getgid();
-	if (setresgid(gid, gid, gid) == -1)
-		err(1, "setresgid");
-}
-#endif
-
 static void
 set_slice_timer(int on)
 {
@@ -309,162 +317,18 @@ clock_gettime_tv(clockid_t clock_id, struct timeval *tv)
 static void
 print_tcp_header(void)
 {
-#ifdef __OpenBSD__
 	char **kv;
-#endif
 
 	if (ptb->rflag == 0)
 		return;
 
 	printf("%12s %14s %12s %8s ", "elapsed_ms", "bytes", "mbps",
 	    "bwidth");
-#ifdef __OpenBSD__
 	for (kv = ptb->kvars;  ptb->kvars != NULL && *kv != NULL; kv++)
 		printf("%s%s", kv != ptb->kvars ? "," : "", *kv);
-#endif
 	printf("\n");
 }
 
-#ifdef __OpenBSD__
-static void
-kget(u_long addr, void *buf, size_t size)
-{
-	if (kvm_read(ptb->kvmh, addr, buf, size) != (ssize_t)size)
-		errx(1, "kvm_read: %s", kvm_geterr(ptb->kvmh));
-}
-
-static u_long
-kfind_tcb(int sock)
-{
-	struct inpcbtable tcbtab;
-	struct inpcb *next, *prev;
-	struct inpcb inpcb, prevpcb;
-	struct tcpcb tcpcb;
-
-	struct sockaddr_storage me, them;
-	socklen_t melen, themlen;
-	struct sockaddr_in *in4;
-	struct sockaddr_in6 *in6;
-	char tmp1[64], tmp2[64];
-	int nretry;
-
-	nretry = 10;
-	melen = themlen = sizeof(struct sockaddr_storage);
-	if (getsockname(sock, (struct sockaddr *)&me, &melen) == -1)
-		err(1, "getsockname");
-	if (getpeername(sock, (struct sockaddr *)&them, &themlen) == -1)
-		err(1, "getpeername");
-	if (me.ss_family != them.ss_family)
-		errx(1, "%s: me.ss_family != them.ss_family", __func__);
-	if (me.ss_family != AF_INET && me.ss_family != AF_INET6)
-		errx(1, "%s: unknown socket family", __func__);
-	if (ptb->vflag >= 2) {
-		saddr_ntop((struct sockaddr *)&me, me.ss_len,
-		    tmp1, sizeof(tmp1));
-		saddr_ntop((struct sockaddr *)&them, them.ss_len,
-		    tmp2, sizeof(tmp2));
-		fprintf(stderr, "Our socket local %s remote %s\n", tmp1, tmp2);
-	}
-	if (ptb->vflag >= 2)
-		fprintf(stderr, "Using PCB table at %lu\n", ptb->ktcbtab);
-retry:
-	kget(ptb->ktcbtab, &tcbtab, sizeof(tcbtab));
-	prev = NULL;
-	next = TAILQ_FIRST(&tcbtab.inpt_queue);
-
-	if (ptb->vflag >= 2)
-		fprintf(stderr, "PCB start at %p\n", next);
-	while (next != NULL) {
-		if (ptb->vflag >= 2)
-			fprintf(stderr, "Checking PCB %p\n", next);
-		kget((u_long)next, &inpcb, sizeof(inpcb));
-		if (prev != NULL) {
-			kget((u_long)prev, &prevpcb, sizeof(prevpcb));
-			if (TAILQ_NEXT(&prevpcb, inp_queue) != next) {
-				if (nretry--) {
-					warnx("PCB prev pointer insane");
-					goto retry;
-				} else
-					errx(1, "PCB prev pointer insane,"
-					    " all attempts exhaused");
-			}
-		}
-		prev = next;
-		next = TAILQ_NEXT(&inpcb, inp_queue);
-
-		if (me.ss_family == AF_INET) {
-			if ((inpcb.inp_flags & INP_IPV6) != 0) {
-				if (ptb->vflag >= 2)
-					fprintf(stderr, "Skip: INP_IPV6");
-				continue;
-			}
-			if (ptb->vflag >= 2) {
-				inet_ntop(AF_INET, &inpcb.inp_laddr,
-				    tmp1, sizeof(tmp1));
-				inet_ntop(AF_INET, &inpcb.inp_faddr,
-				    tmp2, sizeof(tmp2));
-				fprintf(stderr, "PCB %p local: [%s]:%d "
-				    "remote: [%s]:%d\n", prev,
-				    tmp1, inpcb.inp_lport,
-				    tmp2, inpcb.inp_fport);
-			}
-			in4 = (struct sockaddr_in *)&me;
-			if (memcmp(&in4->sin_addr, &inpcb.inp_laddr,
-			    sizeof(struct in_addr)) != 0 ||
-			    in4->sin_port != inpcb.inp_lport)
-				continue;
-			in4 = (struct sockaddr_in *)&them;
-			if (memcmp(&in4->sin_addr, &inpcb.inp_faddr,
-			    sizeof(struct in_addr)) != 0 ||
-			    in4->sin_port != inpcb.inp_fport)
-				continue;
-		} else {
-			if ((inpcb.inp_flags & INP_IPV6) == 0)
-				continue;
-			if (ptb->vflag >= 2) {
-				inet_ntop(AF_INET6, &inpcb.inp_laddr6,
-				    tmp1, sizeof(tmp1));
-				inet_ntop(AF_INET6, &inpcb.inp_faddr6,
-				    tmp2, sizeof(tmp2));
-				fprintf(stderr, "PCB %p local: [%s]:%d "
-				    "remote: [%s]:%d\n", prev,
-				    tmp1, inpcb.inp_lport,
-				    tmp2, inpcb.inp_fport);
-			}
-			in6 = (struct sockaddr_in6 *)&me;
-			if (memcmp(&in6->sin6_addr, &inpcb.inp_laddr6,
-			    sizeof(struct in6_addr)) != 0 ||
-			    in6->sin6_port != inpcb.inp_lport)
-				continue;
-			in6 = (struct sockaddr_in6 *)&them;
-			if (memcmp(&in6->sin6_addr, &inpcb.inp_faddr6,
-			    sizeof(struct in6_addr)) != 0 ||
-			    in6->sin6_port != inpcb.inp_fport)
-				continue;
-		}
-		kget((u_long)inpcb.inp_ppcb, &tcpcb, sizeof(tcpcb));
-		if (tcpcb.t_state != TCPS_ESTABLISHED) {
-			if (ptb->vflag >= 2)
-				fprintf(stderr, "Not established\n");
-			continue;
-		}
-		if (ptb->vflag >= 2)
-			fprintf(stderr, "Found PCB at %p\n", prev);
-		return ((u_long)prev);
-	}
-
-	errx(1, "No matching PCB found");
-}
-
-static void
-kupdate_stats(u_long tcbaddr, struct inpcb *inpcb,
-    struct tcpcb *tcpcb, struct socket *sockb)
-{
-	kget(tcbaddr, inpcb, sizeof(*inpcb));
-	kget((u_long)inpcb->inp_ppcb, tcpcb, sizeof(*tcpcb));
-	kget((u_long)inpcb->inp_socket, sockb, sizeof(*sockb));
-}
-
 static void
 check_kvar(const char *var)
 {
@@ -502,7 +366,6 @@ check_prepare_kvars(char *list)
 	}
 	return (ret);
 }
-#endif
 
 static void
 stats_prepare(struct statctx *sc)
@@ -510,83 +373,122 @@ stats_prepare(struct statctx *sc)
 	sc->buf = ptb->dummybuf;
 	sc->buflen = ptb->dummybuf_len;
 
-#ifdef __OpenBSD__
-	if (ptb->kvars)
-		sc->tcp_tcbaddr = kfind_tcb(sc->fd);
-#endif
 	if (clock_gettime_tv(CLOCK_MONOTONIC, &sc->t_start) == -1)
 		err(1, "clock_gettime_tv");
 	sc->t_last = sc->t_start;
-
+	if (!timerisset(&mainstats.t_first))
+		mainstats.t_first = sc->t_start;
 }
 
-#ifdef __OpenBSD__
 static void
-tcp_stats_display(unsigned long long total_elapsed, long double mbps,
-    float bwperc, struct statctx *sc, struct inpcb *inpcb,
-    struct tcpcb *tcpcb, struct socket *sockb)
-#else
+summary_display(void)
+{
+	struct timeval t_cur, t_diff;
+	long double std_dev;
+	unsigned long long total_elapsed;
+	char *direction;
+
+	if (!ptb->sflag) {
+		direction = "sent";
+		printf("--- %s tcpbench statistics ---\n", mainstats.host);
+	} else {
+		direction = "received";
+		printf("--- tcpbench server statistics ---\n");
+	}
+
+	std_dev = sqrtl(mainstats.nvariance_mbps / mainstats.n_slices);
+
+	if (clock_gettime_tv(CLOCK_MONOTONIC, &t_cur) == -1)
+		err(1, "clock_gettime_tv");
+	timersub(&t_cur, &mainstats.t_first, &t_diff);
+	total_elapsed = t_diff.tv_sec * 1000 + t_diff.tv_usec / 1000;
+
+	printf("%llu bytes %s over %.3Lf seconds\n",
+	    mainstats.total_bytes, direction, total_elapsed/1000.0L);
+	printf("bandwidth min/avg/max/std-dev = %.3Lf/%.3Lf/%.3Lf/%.3Lf Mbps\n",
+	    mainstats.floor_mbps, mainstats.mean_mbps, mainstats.peak_mbps,
+	    std_dev);
+}
+
 static void
 tcp_stats_display(unsigned long long total_elapsed, long double mbps,
-    float bwperc, struct statctx *sc)
-#endif
+    float bwperc, struct statctx *sc, struct tcp_info *tcpi)
 {
-#ifdef __OpenBSD__
 	int j;
-#endif
 
 	printf("%12llu %14llu %12.3Lf %7.2f%% ", total_elapsed, sc->bytes,
 	    mbps, bwperc);
 
-#ifdef __OpenBSD__
 	if (ptb->kvars != NULL) {
-		kupdate_stats(sc->tcp_tcbaddr, inpcb, tcpcb,
-		    sockb);
-
 		for (j = 0; ptb->kvars[j] != NULL; j++) {
 #define S(a) #a
 #define P(b, v, f)							\
-			if (strcmp(ptb->kvars[j], S(b.v)) == 0) {	\
-				printf("%s"f, j > 0 ? "," : "", b->v);	\
+			if (strcmp(ptb->kvars[j], S(v)) == 0) {		\
+				printf("%s"f, j > 0 ? "," : "", b->tcpi_##v); \
 				continue;				\
 			}
-			P(inpcb, inp_flags, "0x%08x")
-			P(sockb, so_rcv.sb_cc, "%lu")
-			P(sockb, so_rcv.sb_hiwat, "%lu")
-			P(sockb, so_rcv.sb_wat, "%lu")
-			P(sockb, so_snd.sb_cc, "%lu")
-			P(sockb, so_snd.sb_hiwat, "%lu")
-			P(sockb, so_snd.sb_wat, "%lu")
-			P(tcpcb, last_ack_sent, "%u")
-			P(tcpcb, max_sndwnd, "%lu")
-			P(tcpcb, rcv_adv, "%u")
-			P(tcpcb, rcv_nxt, "%u")
-			P(tcpcb, rcv_scale, "%u")
-			P(tcpcb, rcv_wnd, "%lu")
-			P(tcpcb, rfbuf_cnt, "%u")
-			P(tcpcb, rfbuf_ts, "%u")
-			P(tcpcb, snd_cwnd, "%lu")
-			P(tcpcb, snd_max, "%u")
-			P(tcpcb, snd_nxt, "%u")
-			P(tcpcb, snd_scale, "%u")
-			P(tcpcb, snd_ssthresh, "%lu")
-			P(tcpcb, snd_una, "%u")
-			P(tcpcb, snd_wl1, "%u")
-			P(tcpcb, snd_wl2, "%u")
-			P(tcpcb, snd_wnd, "%lu")
-			P(tcpcb, t_rcvtime, "%u")
-			P(tcpcb, t_rtseq, "%u")
-			P(tcpcb, t_rttmin, "%hu")
-			P(tcpcb, t_rtttime, "%u")
-			P(tcpcb, t_rttvar, "%hu")
-			P(tcpcb, t_srtt, "%hu")
-			P(tcpcb, ts_recent, "%u")
-			P(tcpcb, ts_recent_age, "%u")
+			P(tcpi, last_ack_recv, "%u")
+			P(tcpi, last_ack_sent, "%u")
+			P(tcpi, last_data_recv, "%u")
+			P(tcpi, last_data_sent, "%u")
+#ifdef __OpenBSD__
+			P(tcpi, max_sndwnd, "%u")
+#endif
+			P(tcpi, options, "%hhu")
+#ifdef __OpenBSD__
+			P(tcpi, rcv_adv, "%u")
+#endif
+			P(tcpi, rcv_mss, "%u")
+#ifdef __OpenBSD__
+			P(tcpi, rcv_nxt, "%u")
+			P(tcpi, rcv_ooopack, "%u")
+#endif
+			P(tcpi, rcv_space, "%u")
+#ifdef __OpenBSD__
+			P(tcpi, rcv_up, "%u")
+#endif
+			P(tcpi, rcv_wscale, "%hhu")
+#ifdef __OpenBSD__
+			P(tcpi, rfbuf_cnt, "%u")
+			P(tcpi, rfbuf_ts, "%u")
+#endif
+			P(tcpi, rtt, "%u")
+#ifdef __OpenBSD__
+			P(tcpi, rttmin, "%u")
+#endif
+			P(tcpi, rttvar, "%u")
+			P(tcpi, snd_cwnd, "%u")
+#ifdef __OpenBSD__
+			P(tcpi, snd_max, "%u")
+#endif
+			P(tcpi, snd_mss, "%u")
+#ifdef __OpenBSD__
+			P(tcpi, snd_nxt, "%u")
+			P(tcpi, snd_rexmitpack, "%u")
+#endif
+			P(tcpi, snd_ssthresh, "%u")
+#ifdef __OpenBSD__
+			P(tcpi, snd_una, "%u")
+			P(tcpi, snd_wl1, "%u")
+			P(tcpi, snd_wl2, "%u")
+			P(tcpi, snd_wnd, "%u")
+			P(tcpi, snd_wscale, "%hhu")
+			P(tcpi, snd_zerowin, "%u")
+			P(tcpi, so_rcv_sb_cc, "%u")
+			P(tcpi, so_rcv_sb_hiwat, "%u")
+			P(tcpi, so_rcv_sb_lowat, "%u")
+			P(tcpi, so_rcv_sb_wat, "%u")
+			P(tcpi, so_snd_sb_cc, "%u")
+			P(tcpi, so_snd_sb_hiwat, "%u")
+			P(tcpi, so_snd_sb_lowat, "%u")
+			P(tcpi, so_snd_sb_wat, "%u")
+			P(tcpi, ts_recent, "%u")
+			P(tcpi, ts_recent_age, "%u")
+#endif
 #undef S
 #undef P
 		}
 	}
-#endif
 	printf("\n");
 }
 
@@ -594,39 +496,39 @@ static void
 tcp_process_slice(int fd, short event, void *bula)
 {
 	unsigned long long total_elapsed, since_last;
-	long double mbps, slice_mbps = 0;
+	long double mbps, old_mean_mbps, slice_mbps = 0;
 	float bwperc;
 	struct statctx *sc;
 	struct timeval t_cur, t_diff;
-#ifdef __OpenBSD__
-	struct inpcb inpcb;
-	struct tcpcb tcpcb;
-	struct socket sockb;
-#endif
+	struct tcp_info tcpi;
+	socklen_t tcpilen;
+
+	if (TAILQ_EMPTY(&sc_queue))
+		return; /* don't pollute stats */
+
+	mainstats.n_slices++;
 
 	TAILQ_FOREACH(sc, &sc_queue, entry) {
 		if (clock_gettime_tv(CLOCK_MONOTONIC, &t_cur) == -1)
 			err(1, "clock_gettime_tv");
-#ifdef __OpenBSD__
-		if (ptb->kvars != NULL) /* process kernel stats */
-			kupdate_stats(sc->tcp_tcbaddr, &inpcb, &tcpcb,
-			    &sockb);
-#endif
+		if (ptb->kvars != NULL) { /* process kernel stats */
+			tcpilen = sizeof(tcpi);
+			if (getsockopt(sc->fd, IPPROTO_TCP, TCP_INFO,
+			    &tcpi, &tcpilen) == -1)
+				err(1, "get tcp_info");
+		}
 
 		timersub(&t_cur, &sc->t_start, &t_diff);
 		total_elapsed = t_diff.tv_sec * 1000 + t_diff.tv_usec / 1000;
 		timersub(&t_cur, &sc->t_last, &t_diff);
 		since_last = t_diff.tv_sec * 1000 + t_diff.tv_usec / 1000;
+		if (since_last == 0)
+			continue;
 		bwperc = (sc->bytes * 100.0) / mainstats.slice_bytes;
 		mbps = (sc->bytes * 8) / (since_last * 1000.0);
 		slice_mbps += mbps;
 
-#ifdef __OpenBSD__
-		tcp_stats_display(total_elapsed, mbps, bwperc, sc,
-		    &inpcb, &tcpcb, &sockb);
-#else
-		tcp_stats_display(total_elapsed, mbps, bwperc, sc);
-#endif
+		tcp_stats_display(total_elapsed, mbps, bwperc, sc, &tcpi);
 
 		sc->t_last = t_cur;
 		sc->bytes = 0;
@@ -635,49 +537,79 @@ tcp_process_slice(int fd, short event, void *bula)
 	/* process stats for this slice */
 	if (slice_mbps > mainstats.peak_mbps)
 		mainstats.peak_mbps = slice_mbps;
+	if (slice_mbps < mainstats.floor_mbps)
+		mainstats.floor_mbps = slice_mbps;
+	old_mean_mbps = mainstats.mean_mbps;
+	mainstats.mean_mbps += (slice_mbps - mainstats.mean_mbps) /
+				mainstats.n_slices;
+
+	/* "Welford's method" for online variance
+	 * see Knuth, TAoCP Volume 2, 3rd edn., p232 */
+	mainstats.nvariance_mbps += (slice_mbps - old_mean_mbps) *
+				    (slice_mbps - mainstats.mean_mbps);
+
 	printf("Conn: %3d Mbps: %12.3Lf Peak Mbps: %12.3Lf Avg Mbps: %12.3Lf\n",
 	    mainstats.nconns, slice_mbps, mainstats.peak_mbps,
 	    mainstats.nconns ? slice_mbps / mainstats.nconns : 0);
-	mainstats.slice_bytes = 0;
 
+	mainstats.slice_bytes = 0;
 	set_slice_timer(mainstats.nconns > 0);
 }
 
 static void
-udp_process_slice(int fd, short event, void *v_sc)
+udp_process_slice(int fd, short event, void *bula)
 {
-	struct statctx *sc = v_sc;
 	unsigned long long total_elapsed, since_last, pps;
-	long double slice_mbps;
+	long double old_mean_mbps, slice_mbps;
 	struct timeval t_cur, t_diff;
 
+	mainstats.n_slices++;
+
 	if (clock_gettime_tv(CLOCK_MONOTONIC, &t_cur) == -1)
 		err(1, "clock_gettime_tv");
-	/* Calculate pps */
-	timersub(&t_cur, &sc->t_start, &t_diff);
+
+	timersub(&t_cur, &udp_sc->t_start, &t_diff);
 	total_elapsed = t_diff.tv_sec * 1000 + t_diff.tv_usec / 1000;
-	timersub(&t_cur, &sc->t_last, &t_diff);
+
+	timersub(&t_cur, &udp_sc->t_last, &t_diff);
 	since_last = t_diff.tv_sec * 1000 + t_diff.tv_usec / 1000;
-	slice_mbps = (sc->bytes * 8) / (since_last * 1000.0);
-	pps = (sc->udp_slice_pkts * 1000) / since_last;
+	if (since_last == 0)
+		return;
+
+	slice_mbps = (udp_sc->bytes * 8) / (since_last * 1000.0);
+	pps = (udp_sc->udp_slice_pkts * 1000) / since_last;
+
 	if (slice_mbps > mainstats.peak_mbps)
 		mainstats.peak_mbps = slice_mbps;
+	if (slice_mbps < mainstats.floor_mbps)
+		mainstats.floor_mbps = slice_mbps;
+	old_mean_mbps = mainstats.mean_mbps;
+	mainstats.mean_mbps += (slice_mbps - mainstats.mean_mbps) /
+				mainstats.n_slices;
+
+	/* "Welford's method" for online variance
+	 * see Knuth, TAoCP Volume 2, 3rd edn., p232 */
+	mainstats.nvariance_mbps += (slice_mbps - old_mean_mbps) *
+				    (slice_mbps - mainstats.mean_mbps);
+
 	printf("Elapsed: %11llu Mbps: %11.3Lf Peak Mbps: %11.3Lf %s PPS: %7llu\n",
 	    total_elapsed, slice_mbps, mainstats.peak_mbps,
 	    ptb->sflag ? "Rx" : "Tx", pps);
 
 	/* Clean up this slice time */
-	sc->t_last = t_cur;
-	sc->bytes = 0;
-	sc->udp_slice_pkts = 0;
+	udp_sc->t_last = t_cur;
+	udp_sc->bytes = 0;
+	udp_sc->udp_slice_pkts = 0;
+
+	mainstats.slice_bytes = 0;
 	set_slice_timer(1);
 }
 
 static void
-udp_server_handle_sc(int fd, short event, void *v_sc)
+udp_server_handle_sc(int fd, short event, void *bula)
 {
+	static int first_read = 1;
 	ssize_t n;
-	struct statctx *sc = v_sc;
 
 	n = read(fd, ptb->dummybuf, ptb->dummybuf_len);
 	if (n == 0)
@@ -690,12 +622,16 @@ udp_server_handle_sc(int fd, short event, void *v_sc)
 
 	if (ptb->vflag >= 3)
 		fprintf(stderr, "read: %zd bytes\n", n);
-	/* If this was our first packet, start slice timer */
-	if (mainstats.peak_mbps == 0)
+	if (first_read) {
+		first_read = 0;
+		stats_prepare(udp_sc);
 		set_slice_timer(1);
+	}
 	/* Account packet */
-	sc->udp_slice_pkts++;
-	sc->bytes += n;
+	udp_sc->udp_slice_pkts++;
+	udp_sc->bytes += n;
+	mainstats.slice_bytes += n;
+	mainstats.total_bytes += n;
 }
 
 static void
@@ -732,6 +668,7 @@ tcp_server_handle_sc(int fd, short event, void *v_sc)
 		fprintf(stderr, "read: %zd bytes\n", n);
 	sc->bytes += n;
 	mainstats.slice_bytes += n;
+	mainstats.total_bytes += n;
 }
 
 static void
@@ -742,7 +679,7 @@ tcp_server_accept(int fd, short event, void *arg)
 	struct statctx *sc;
 	struct sockaddr_storage ss;
 	socklen_t sslen;
-	char tmp[128];
+	char tmp[NI_MAXHOST + 2 + NI_MAXSERV];
 
 	sslen = sizeof(ss);
 
@@ -783,10 +720,12 @@ tcp_server_accept(int fd, short event, void *arg)
 	sc->tcp_ts = ts;
 	sc->fd = sock;
 	stats_prepare(sc);
+
 	event_set(&sc->ev, sc->fd, EV_READ | EV_PERSIST,
 	    tcp_server_handle_sc, sc);
 	event_add(&sc->ev, NULL);
 	TAILQ_INSERT_TAIL(&sc_queue, sc, entry);
+
 	mainstats.nconns++;
 	if (mainstats.nconns == 1)
 		set_slice_timer(1);
@@ -796,9 +735,8 @@ tcp_server_accept(int fd, short event, void *arg)
 }
 
 static void
-server_init(struct addrinfo *aitop, struct statctx *udp_sc)
+server_init(struct addrinfo *aitop)
 {
-	char tmp[128];
 	int sock, on = 1;
 	struct addrinfo *ai;
 	struct event *ev;
@@ -807,6 +745,8 @@ server_init(struct addrinfo *aitop, struct statctx *udp_sc)
 
 	lnfds = 0;
 	for (ai = aitop; ai != NULL; ai = ai->ai_next) {
+		char tmp[NI_MAXHOST + 2 + NI_MAXSERV];
+
 		saddr_ntop(ai->ai_addr, ai->ai_addrlen, tmp, sizeof(tmp));
 		if (ptb->vflag)
 			fprintf(stderr, "Try to bind to %s\n", tmp);
@@ -818,6 +758,11 @@ server_init(struct addrinfo *aitop, struct statctx *udp_sc)
 				warn("socket");
 			continue;
 		}
+		if (ptb->Dflag) {
+			if (setsockopt(sock, SOL_SOCKET, SO_DEBUG,
+			    &ptb->Dflag, sizeof(ptb->Dflag)))
+				err(1, "setsockopt SO_DEBUG");
+		}
 		if (ptb->Tflag != -1 && ai->ai_family == AF_INET) {
 			if (setsockopt(sock, IPPROTO_IP, IP_TOS,
 			    &ptb->Tflag, sizeof(ptb->Tflag)))
@@ -828,6 +773,13 @@ server_init(struct addrinfo *aitop, struct statctx *udp_sc)
 			    &ptb->Tflag, sizeof(ptb->Tflag)))
 				err(1, "setsockopt IPV6_TCLASS");
 		}
+#ifndef __OpenBSD__
+		if (ai->ai_family == AF_INET6) {
+			if (setsockopt(sock, IPPROTO_IPV6, IPV6_V6ONLY,
+			    &on, sizeof(on)) == -1)
+				warn("ipv6 only");
+		}
+#endif
 		if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
 		    &on, sizeof(on)) == -1)
 			warn("reuse port");
@@ -858,7 +810,7 @@ server_init(struct addrinfo *aitop, struct statctx *udp_sc)
 			if ((ev = calloc(1, sizeof(*ev))) == NULL)
 				err(1, "calloc");
 			event_set(ev, sock, EV_READ | EV_PERSIST,
-			    udp_server_handle_sc, udp_sc);
+			    udp_server_handle_sc, NULL);
 			event_add(ev, NULL);
 		} else {
 			if ((ts = calloc(1, sizeof(*ts))) == NULL)
@@ -893,32 +845,33 @@ client_handle_sc(int fd, short event, void *v_sc)
 		if (errno == EINTR || errno == EWOULDBLOCK ||
 		    (UDP_MODE && errno == ENOBUFS))
 			return;
-		err(1, "write");
+		warn("write");
+		wrapup(1);
 	}
 	if (TCP_MODE && n == 0) {
 		fprintf(stderr, "Remote end closed connection");
-		exit(1);
+		wrapup(1);
 	}
 	if (ptb->vflag >= 3)
 		fprintf(stderr, "write: %zd bytes\n", n);
 	sc->bytes += n;
 	mainstats.slice_bytes += n;
+	mainstats.total_bytes += n;
 	if (UDP_MODE)
 		sc->udp_slice_pkts++;
 }
 
 static void
-client_init(struct addrinfo *aitop, int nconn, struct statctx *udp_sc,
-    struct addrinfo *aib)
+client_init(struct addrinfo *aitop, int nconn, struct addrinfo *aib)
 {
 	struct statctx *sc;
 	struct addrinfo *ai;
-	char tmp[128];
 	int i, r, sock;
 
-	sc = udp_sc;
 	for (i = 0; i < nconn; i++) {
 		for (sock = -1, ai = aitop; ai != NULL; ai = ai->ai_next) {
+			char tmp[NI_MAXHOST + 2 + NI_MAXSERV];
+
 			saddr_ntop(ai->ai_addr, ai->ai_addrlen, tmp,
 			    sizeof(tmp));
 			if (ptb->vflag && i == 0)
@@ -931,6 +884,11 @@ client_init(struct addrinfo *aitop, int nconn, struct statctx *udp_sc,
 					warn("socket");
 				continue;
 			}
+			if (ptb->Dflag) {
+				if (setsockopt(sock, SOL_SOCKET, SO_DEBUG,
+				    &ptb->Dflag, sizeof(ptb->Dflag)))
+					err(1, "setsockopt SO_DEBUG");
+			}
 			if (aib != NULL) {
 				saddr_ntop(aib->ai_addr, aib->ai_addrlen,
 				    tmp, sizeof(tmp));
@@ -978,13 +936,17 @@ client_init(struct addrinfo *aitop, int nconn, struct statctx *udp_sc,
 		if (TCP_MODE) {
 			if ((sc = calloc(1, sizeof(*sc))) == NULL)
 				err(1, "calloc");
-		}
+		} else
+			sc = udp_sc;
+
 		sc->fd = sock;
 		stats_prepare(sc);
+
 		event_set(&sc->ev, sc->fd, EV_WRITE | EV_PERSIST,
 		    client_handle_sc, sc);
 		event_add(&sc->ev, NULL);
 		TAILQ_INSERT_TAIL(&sc_queue, sc, entry);
+
 		mainstats.nconns++;
 		if (mainstats.nconns == 1)
 			set_slice_timer(1);
@@ -1052,7 +1014,28 @@ map_tos(char *s, int *val)
 static void
 quit(int sig, short event, void *arg)
 {
-	exit(0);
+	wrapup(0);
+}
+
+static void
+wrapup(int err)
+{
+	const int transfers = timerisset(&mainstats.t_first);
+	const int stats = (mainstats.floor_mbps != INFINITY);
+
+	if (transfers) {
+		if (!stats) {
+			if (UDP_MODE)
+				udp_process_slice(0, 0, NULL);
+			else
+				tcp_process_slice(0, 0, NULL);
+		}
+
+		summary_display();
+	}
+
+	if (err != -1)
+		exit(err);
 }
 
 int
@@ -1061,39 +1044,38 @@ main(int argc, char **argv)
 	struct timeval tv;
 #ifdef __OpenBSD__
 	unsigned int secs, rtable;
-	char kerr[_POSIX2_LINE_MAX], *tmp;
 #else
 	unsigned int secs;
 #endif
+	char *tmp;
 	struct addrinfo *aitop, *aib, hints;
 	const char *errstr;
 	struct rlimit rl;
 	int ch, herr, nconn;
 	int family = PF_UNSPEC;
-#ifdef __OpenBSD__
-	struct nlist nl[] = { { "_tcbtable" }, { "" } };
-#endif
 	const char *host = NULL, *port = DEFAULT_PORT, *srcbind = NULL;
+#ifdef __OpenBSD__
+	struct event ev_sigint, ev_sigterm, ev_sighup, ev_siginfo, ev_progtimer;
+#else
 	struct event ev_sigint, ev_sigterm, ev_sighup, ev_progtimer;
-	struct statctx *udp_sc = NULL;
+#endif
 	struct sockaddr_un sock_un;
 
 	/* Init world */
 	setvbuf(stdout, NULL, _IOLBF, 0);
 	ptb = &tcpbench;
 	ptb->dummybuf_len = 0;
+	ptb->Dflag = 0;
 	ptb->Sflag = ptb->sflag = ptb->vflag = ptb->Rflag = ptb->Uflag = 0;
-#ifdef __OpenBSD__
-	ptb->kvmh  = NULL;
 	ptb->kvars = NULL;
-#endif
 	ptb->rflag = DEFAULT_STATS_INTERVAL;
 	ptb->Tflag = -1;
 	nconn = 1;
 	aib = NULL;
 	secs = 0;
 
-	while ((ch = getopt(argc, argv, "46b:B:hlk:n:p:Rr:sS:t:T:uUvV:")) != -1) {
+	while ((ch = getopt(argc, argv, "46b:B:Dhlk:n:p:Rr:sS:t:T:uUvV:"))
+	    != -1) {
 		switch (ch) {
 		case '4':
 			family = PF_INET;
@@ -1104,7 +1086,9 @@ main(int argc, char **argv)
 		case 'b':
 			srcbind = optarg;
 			break;
-#ifdef __OpenBSD__
+		case 'D':
+			ptb->Dflag = 1;
+			break;
 		case 'l':
 			list_kvars();
 			exit(0);
@@ -1114,7 +1098,6 @@ main(int argc, char **argv)
 			ptb->kvars = check_prepare_kvars(tmp);
 			free(tmp);
 			break;
-#endif
 		case 'R':
 			ptb->Rflag = 1;
 			break;
@@ -1199,52 +1182,23 @@ main(int argc, char **argv)
 	}
 
 #ifdef __OpenBSD__
-	if (pledge("stdio unveil rpath dns inet unix id proc", NULL) == -1)
+	if (pledge("stdio unveil rpath dns inet unix id", NULL) == -1)
 		err(1, "pledge");
 #endif
 
 	argv += optind;
 	argc -= optind;
-#ifdef __OpenBSD__
 	if ((argc != (ptb->sflag && !ptb->Uflag ? 0 : 1)) ||
 	    (UDP_MODE && (ptb->kvars || nconn != 1)))
-#else
-	if ((argc != (ptb->sflag ? 0 : 1)) ||
-	    (UDP_MODE && (nconn != 1)))
-#endif
 		usage();
 
-#ifdef __OpenBSD__
-	if (ptb->kvars) {
-		if (unveil(_PATH_MEM, "r") == -1)
-			err(1, "unveil");
-		if (unveil(_PATH_KMEM, "r") == -1)
-			err(1, "unveil");
-		if (unveil(_PATH_KSYMS, "r") == -1)
-			err(1, "unveil");
-
-		if ((ptb->kvmh = kvm_openfiles(NULL, NULL, NULL,
-		    O_RDONLY, kerr)) == NULL)
-			errx(1, "kvm_open: %s", kerr);
-		drop_gid();
-		if (kvm_nlist(ptb->kvmh, nl) < 0 || nl[0].n_type == 0)
-			errx(1, "kvm: no namelist");
-		ptb->ktcbtab = nl[0].n_value;
-	} else
-		drop_gid();
-#endif
-
-#ifdef __OpenBSD__
 	if (!ptb->sflag || ptb->Uflag)
-#else
-	if (!ptb->sflag)
-#endif
-		host = argv[0];
+		mainstats.host = host = argv[0];
 
 #ifdef __OpenBSD__
 	if (ptb->Uflag)
 		if (unveil(host, "rwc") == -1)
-			err(1, "unveil");
+			err(1, "unveil %s", host);
 
 	if (pledge("stdio id dns inet unix", NULL) == -1)
 		err(1, "pledge");
@@ -1330,29 +1284,37 @@ main(int argc, char **argv)
 		err(1, "malloc");
 	arc4random_buf(ptb->dummybuf, ptb->dummybuf_len);
 
+	timerclear(&mainstats.t_first);
+	mainstats.floor_mbps = INFINITY;
+
 	/* Setup libevent and signals */
 	event_init();
 	signal_set(&ev_sigterm, SIGTERM, signal_handler, NULL);
 	signal_set(&ev_sighup, SIGHUP, signal_handler, NULL);
 	signal_set(&ev_sigint, SIGINT, signal_handler, NULL);
+#ifdef __OpenBSD__
+	signal_set(&ev_siginfo, SIGINFO, signal_handler, NULL);
+#endif
 	signal_add(&ev_sigint, NULL);
 	signal_add(&ev_sigterm, NULL);
 	signal_add(&ev_sighup, NULL);
+#ifdef __OpenBSD__
+	signal_add(&ev_siginfo, NULL);
+#endif
 	signal(SIGPIPE, SIG_IGN);
 
 	if (UDP_MODE) {
 		if ((udp_sc = calloc(1, sizeof(*udp_sc))) == NULL)
 			err(1, "calloc");
 		udp_sc->fd = -1;
-		stats_prepare(udp_sc);
-		evtimer_set(&mainstats.timer, udp_process_slice, udp_sc);
+		evtimer_set(&mainstats.timer, udp_process_slice, NULL);
 	} else {
 		print_tcp_header();
 		evtimer_set(&mainstats.timer, tcp_process_slice, NULL);
 	}
 
 	if (ptb->sflag)
-		server_init(aitop, udp_sc);
+		server_init(aitop);
 	else {
 		if (secs > 0) {
 			timerclear(&tv);
@@ -1360,10 +1322,10 @@ main(int argc, char **argv)
 			evtimer_set(&ev_progtimer, quit, NULL);
 			evtimer_add(&ev_progtimer, &tv);
 		}
-		client_init(aitop, nconn, udp_sc, aib);
+		client_init(aitop, nconn, aib);
 
 #ifdef __OpenBSD__
-		if (pledge("stdio", NULL) == -1)
+		if (pledge("stdio inet", NULL) == -1)
 			err(1, "pledge");
 #endif
 	}
